/*
nrgrep: grep replacement for logcheck ignores repeating log lines
Copyright (C) 2000 Ed Cashin

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
#include	<stdlib.h>
#include	<regex.h>	/* this gets POSIX features after
				 * we include stdlib.h because
				 * _POSIX_C_SOURCE is defined.
				 */
#include	<unistd.h>
#include	<stdio.h>
#include	<ctype.h>
#include	<string.h>
#include	<errno.h>
#include	"xmalloc.h"	/* from elc library */
#include	"vector.h"	/* from elc library */
#include	"match.h"

/* macro to retrieve the pointer to char pointer at index i from vector v. 
   you still have to dereference it to get a char pointer. */
#define	VEC_AT(v, i)	((char **) (v->index(v, i)))

/* The regular expression that should match the log message that your
   system uses for repeating messages. 

   You can override this default, which works on my Linux 6.1 systems,
   in the makefile.
*/
#ifndef REPEAT_REGEX
#define	REPEAT_REGEX	"last message repeated [0-9]+ times"
#endif

/* file static global variables */
static int nrgrep_was_output	 = 0;

void usage(const char *progname)
{
    printf("%s --- special grep for logcheck can ignore repeating messages\n",
	   progname);
    printf("\nusage:\n\t%s -v -f ignorefile\n",
	   progname
	   );
    printf("\t  -h\t\tdisplay this help\n");
    printf("\t  -f\t\tignorefile name\n");
    printf("\t  -v\t\tthis option is ignored (it's always -v)\n");
}

void chomp(char *chp)
{
    /* taken from b_day.  I use this a lot.
       Fri Feb 18 13:11:15 EST 2000
     */
    for( ; *chp; ++chp)
      if(*chp == '\n')
	*chp = '\0';
}

void load_ignorables(vec_t *ignorables, const char *ignorefile)
{
    char	buf[BUFSIZ];
    char	*line	 = NULL;
    FILE	*input	 = fopen(ignorefile, "r");
    
    while (fgets(buf, BUFSIZ, input)) {
      line	 = xstrdup(buf);
      chomp(line);
      if ((line[0] != '\0') && (line[0] != '#'))
	ignorables->add(ignorables, &line);
    }

    fclose(input);
}

void norepeato_grep(vec_t *ignorables, FILE *input)
{
    char	buf[BUFSIZ];
    int		i;
    int		end		 = ignorables->n_elems;
#ifdef	NO_INITIAL_REPEATS
    int		ignore_line	 = 1; /* ignore leading
				       * "last message repeated"
				       * lines by setting this to one */
#else
    int		ignore_line	 = 0;
#endif

    while(fgets(buf, BUFSIZ, input)){

      /* here's the non-repeato part: */
      if (ignore_line
	  && match(buf, REPEAT_REGEX)) {
	continue;
      } else {
	ignore_line	 = 0;
      }

      for (i = 0; i < end; ++i) {
	if (match(buf, * (VEC_AT(ignorables, i)))) {
	  ignore_line	 = 1;
	  break;
	}
      }

      if (! ignore_line) {
	if (! nrgrep_was_output)
	  nrgrep_was_output	 = 1;
	fputs(buf, stdout);
      }
    }
}

void free_ignorable_regexes(vec_t *v)
{
    /* free all the strings in the vector */
    char	*cp;
    int		i;
    int		end	 = v->n_elems;
    for (i = 0; i < end; ++i) {
      cp	 = * (VEC_AT(v, i));
      xfree(cp);
    }
}

int main(int argc, char *argv[])
{
    int 	c;
    char	*ignorefile	 = NULL;

    /* this is a vector (growable array) that holds pointers to
       char pointers.

       the strings pointed to by the char pointers are the regular
       expressions from the ignorefile.
     */
    vec_t	*ignorables	 = new_vec(sizeof(char *));
    char	*progname	 = strrchr(argv[0], '/');
    FILE	*input		 = NULL;

    if (! progname)
      progname	 = argv[0];
    else
      ++progname;

    opterr = 0;
    while ((c = getopt(argc, argv, "hf:v")) != -1) {
      switch (c) {
	case 'h':
	  usage(progname);
	  exit(EXIT_SUCCESS);
	  break;
	case 'f':
	  ignorefile	 = xstrdup(optarg);
	  break;
	case 'v':		/* ignore -v */
	  break;
	case '?':
	  if (isprint (optopt))
	    fprintf (stderr, "Unknown option `-%c'.  try \"%s -h\"\n",
		     optopt, progname);
	  else
	    fprintf (stderr,
		     "Unknown option character `\\x%x'.  try \"%s -h\"\n",
		     optopt, progname);
	  exit(EXIT_FAILURE);
	  break;
	default:
	  /* control shouldn't reach here. */
	  usage(progname);
	  abort ();
      }
    }

    if(!(optind < argc)){
      input	 = stdin;
    } else {
      input	 = fopen(argv[optind], "r");
      if (! input) {
	fprintf(stderr,
		"Error: could not open %s for reading: %s\n",
		argv[optind],
		strerror(errno));
	exit(EXIT_FAILURE);
      }
    }
    
    if (! ignorefile) {
      fprintf(stderr, "Error: no ignorefile specified\n\n");
      usage(progname);
      exit(EXIT_FAILURE);
    }

    load_ignorables(ignorables, ignorefile);
    norepeato_grep(ignorables, input);

    xfree(ignorefile);
    free_ignorable_regexes(ignorables);
    ignorables->destroy(ignorables);
    xfree(ignorables);
    
    /* This is not really standard C, but the grep man page describes
       this behavior:
     */
    return ! nrgrep_was_output;
}

